from typing import Any, Dict

from biolib._index.types import IndexInfo
from biolib.api import client as api_client
from biolib.biolib_api_client.biolib_app_api import _get_app_uri_from_str


def get_index_from_uri(uri: str) -> IndexInfo:
    normalized_uri = _get_app_uri_from_str(uri)
    app_response: Dict[str, Any] = api_client.get(path='/app/', params={'uri': normalized_uri}).json()
    resource_uri = app_response['app_version']['app_uri']
    if app_response['app']['type'] != 'index':
        raise Exception(f'Resource "{resource_uri}" is not an Index')
    return IndexInfo(
        resource_uri=app_response['app_version']['app_uri'],
        resource_uuid=app_response['app']['public_id'],
        group_uuid=app_response['app']['group_uuid'],
    )



import click

from biolib._internal.add_copilot_prompts import add_copilot_prompts


@click.group(name='sdk', help='Advanced commands for developers')
def sdk():
    pass


@sdk.command(
    name='add-copilot-prompts', help='Add BioLib-specific GitHub Copilot prompts and instructions to your repository'
)
@click.option('--force', is_flag=True, help='Force overwrite existing files.')
def add_copilot_prompts_command(force: bool) -> None:
    add_copilot_prompts(force)



import io
import os
import tarfile
import threading
import time

from docker.models.containers import Container

from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.compute_node.utils import BIOLIB_PROXY_NETWORK_NAME
from biolib.compute_node.webserver.proxy_utils import get_biolib_nginx_proxy_image
from biolib.typing_utils import Optional


class LogStreamingThread(threading.Thread):
    def __init__(self, container, container_name: str):
        super().__init__(daemon=True)
        self._container = container
        self._container_name = container_name
        self._stop_event = threading.Event()

    def run(self) -> None:
        try:
            logger_no_user_data.debug(f'Starting log streaming for container "{self._container_name}"')
            log_stream = self._container.logs(follow=True, stream=True)
            for log_line in log_stream:
                if self._stop_event.is_set():
                    break
                if log_line:
                    logger.debug(f'ComputeNodeResultsProxy | {log_line.decode("utf-8", errors="replace").rstrip()}')
        except Exception as error:
            logger_no_user_data.debug(f'Log streaming for container "{self._container_name}" ended: {error}')

    def stop(self) -> None:
        self._stop_event.set()


class ComputeNodeResultsProxy:
    _instance: Optional['ComputeNodeResultsProxy'] = None

    def __init__(self, tls_pem_certificate_path: str, tls_pem_key_path: str):
        assert tls_pem_certificate_path, 'tls_pem_certificate_path is required'
        assert tls_pem_key_path, 'tls_pem_key_path is required'
        self._name = 'biolib-compute-node-results-proxy'
        self._container: Optional[Container] = None
        self._docker = BiolibDockerClient().get_docker_client()
        self._tls_pem_certificate_path = tls_pem_certificate_path
        self._tls_pem_key_path = tls_pem_key_path
        self._log_streaming_thread: Optional[LogStreamingThread] = None

    @staticmethod
    def start_proxy(tls_pem_certificate_path: str, tls_pem_key_path: str) -> None:
        abs_cert_path = os.path.abspath(tls_pem_certificate_path)
        abs_key_path = os.path.abspath(tls_pem_key_path)
        assert os.path.exists(abs_cert_path), f'TLS certificate file does not exist: {abs_cert_path}'
        assert os.path.exists(abs_key_path), f'TLS key file does not exist: {abs_key_path}'

        if ComputeNodeResultsProxy._instance is None:
            logger_no_user_data.debug(
                f'Creating ComputeNodeResultsProxy instance with cert: {abs_cert_path}, key: {abs_key_path}'
            )
            ComputeNodeResultsProxy._instance = ComputeNodeResultsProxy(abs_cert_path, abs_key_path)
            ComputeNodeResultsProxy._instance._start()  # pylint: disable=protected-access

    @staticmethod
    def stop_proxy() -> None:
        if ComputeNodeResultsProxy._instance is not None:
            ComputeNodeResultsProxy._instance._terminate()  # pylint: disable=protected-access
            ComputeNodeResultsProxy._instance = None

    def _start(self) -> None:
        docker = BiolibDockerClient.get_docker_client()

        for index in range(3):
            logger_no_user_data.debug(
                f'Attempt {index} at creating ComputeNodeResultsProxy container "{self._name}"...'
            )
            try:
                self._container = docker.containers.create(
                    detach=True,
                    image=get_biolib_nginx_proxy_image(),
                    name=self._name,
                    network=BIOLIB_PROXY_NETWORK_NAME,
                    ports={'443/tcp': 20443},
                    volumes={
                        self._tls_pem_certificate_path: {'bind': '/etc/ssl/certs/certificate.pem', 'mode': 'ro'},
                        self._tls_pem_key_path: {'bind': '/etc/ssl/private/private_key.pem', 'mode': 'ro'},
                    },
                )
                break
            except Exception as error:
                logger_no_user_data.exception(f'Failed to create container "{self._name}" hit error: {error}')

            logger_no_user_data.debug('Sleeping before re-trying container creation...')
            time.sleep(3)

        if not self._container or not self._container.id:
            raise BioLibError(f'Exceeded re-try limit for creating container {self._name}')

        self._write_nginx_config_to_container()
        self._container.start()

        logger_no_user_data.debug(f'Waiting for container "{self._name}" to be ready...')
        proxy_is_ready = False
        for retry_count in range(1, 5):
            time.sleep(0.5 * retry_count)
            container_logs = self._container.logs()
            if b'ready for start up\n' in container_logs or b'start worker process ' in container_logs:
                proxy_is_ready = True
                break

        if not proxy_is_ready:
            logger_no_user_data.error('ComputeNodeResultsProxy did not start properly.')
            self._terminate()
            raise Exception('ComputeNodeResultsProxy did not start properly')

        self._container.reload()
        logger.debug(f'ComputeNodeResultsProxy container "{self._name}" started with ID: {self._container.id}')

        self._log_streaming_thread = LogStreamingThread(self._container, self._name)
        self._log_streaming_thread.start()
        logger_no_user_data.debug(f'Started log streaming for container "{self._name}"')

    def _terminate(self):
        logger_no_user_data.debug(f'Terminating ComputeNodeResultsProxy container "{self._name}"')
        if self._log_streaming_thread:
            self._log_streaming_thread.stop()
            self._log_streaming_thread = None

        logger.debug(f'Docker container removal temporarily disabled for debugging purposes (container "{self._name}")')
        # TODO: Figure if we need to remove the container or keep it for debugging purposes
        # if self._container:
        #     self._container.remove(force=True)

    def _write_nginx_config_to_container(self) -> None:
        if not self._container:
            raise Exception('ComputeNodeResultsProxy container not defined when attempting to write NGINX config')

        docker = BiolibDockerClient.get_docker_client()

        nginx_config = """
events {
  worker_connections  1024;
}

http {
    client_max_body_size 0;
    resolver 127.0.0.11 ipv6=off valid=30s;

    server {
        listen 443 ssl http2 default_server;

        ssl_certificate /etc/ssl/certs/certificate.pem;
        ssl_certificate_key /etc/ssl/private/private_key.pem;
        ssl_protocols TLSv1.2 TLSv1.3;

        location / {
            if ($http_biolib_result_uuid = "") {
                return 400 "Missing biolib-result-uuid header";
            }

            if ($http_biolib_result_port = "") {
                return 400 "Missing biolib-result-port header";
            }

            set $target_host "biolib-app-caller-proxy-$http_biolib_result_uuid";
            proxy_pass http://$target_host:1080$request_uri;
            proxy_set_header biolib-result-uuid $http_biolib_result_uuid;
            proxy_set_header biolib-result-port $http_biolib_result_port;
            proxy_pass_request_headers on;
        }
    }
}
"""

        nginx_config_bytes = nginx_config.encode()
        tarfile_in_memory = io.BytesIO()
        with tarfile.open(fileobj=tarfile_in_memory, mode='w:gz') as tar:
            info = tarfile.TarInfo('/nginx.conf')
            info.size = len(nginx_config_bytes)
            tar.addfile(info, io.BytesIO(nginx_config_bytes))

        tarfile_bytes = tarfile_in_memory.getvalue()
        tarfile_in_memory.close()
        logger_no_user_data.debug('Writing NGINX configuration to ComputeNodeResultsProxy container')
        docker.api.put_archive(self._container.id, '/etc/nginx', tarfile_bytes)



import time
import uuid
from fnmatch import fnmatch

from biolib.biolib_binary_format.utils import LazyLoadedFile
from biolib.typing_utils import Callable, List, Union, cast

PathFilter = Union[str, Callable[[str], bool]]


def filter_lazy_loaded_files(files: List[LazyLoadedFile], path_filter: PathFilter) -> List[LazyLoadedFile]:
    if not (isinstance(path_filter, str) or callable(path_filter)):
        raise Exception('Expected path_filter to be a string or a function')

    if callable(path_filter):
        return list(filter(lambda x: path_filter(x.path), files))  # type: ignore

    glob_filter = cast(str, path_filter)

    # since all file paths start with /, make sure filter does too
    if not glob_filter.startswith('/'):
        glob_filter = '/' + glob_filter

    def _filter_function(file: LazyLoadedFile) -> bool:
        return fnmatch(file.path, glob_filter)

    return list(filter(_filter_function, files))


def open_browser_window_from_notebook(url_to_open: str) -> None:
    try:
        from IPython.display import (  # type:ignore # pylint: disable=import-error, import-outside-toplevel
            Javascript,
            display,
            update_display,
        )
    except ImportError as error:
        raise Exception('Unexpected environment. This function can only be called from a notebook.') from error

    display_id = str(uuid.uuid4())
    display(Javascript(f'window.open("{url_to_open}");'), display_id=display_id)
    time.sleep(1)
    update_display(Javascript(''), display_id=display_id)



import os
import subprocess
from datetime import datetime, timedelta, timezone

from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.job_worker.cache_state import LfsCacheState


def prune_lfs_cache(dry_run: bool) -> None:
    logger_no_user_data.info(f'Pruning LFS cache (dry run = {dry_run})...')

    current_time = datetime.now(timezone.utc)
    paths_to_delete = set()

    with LfsCacheState() as state:
        lfs_storage_path = f'{LfsCacheState().main_lfs_storage_path}/lfs'
        if os.path.exists(lfs_storage_path):
            for lfs_uuid in os.listdir(lfs_storage_path):
                if lfs_uuid not in state['large_file_systems']:
                    path_to_delete = f'{lfs_storage_path}/{lfs_uuid}'
                    logger_no_user_data.info(f'Will delete path "{path_to_delete}" as it is not in cache state')
                    paths_to_delete.add(path_to_delete)

        lfs_uuids_to_keep_in_state = set()
        for lfs_uuid, lfs in state['large_file_systems'].items():
            last_used_at = datetime.fromisoformat(lfs['last_used_at'])
            if last_used_at.tzinfo is None:
                last_used_at = last_used_at.replace(tzinfo=timezone.utc)
            lfs_time_to_live_in_days = 60 if lfs['state'] == 'ready' else 7

            if last_used_at < current_time - timedelta(days=lfs_time_to_live_in_days):
                logger_no_user_data.info(f"Will delete LFS {lfs_uuid} as it was last used at {lfs['last_used_at']}")
                partition = state['storage_partitions'][lfs['storage_partition_uuid']]
                paths_to_delete.add(f"{partition['path']}/lfs/{lfs_uuid}")
                if not dry_run:
                    partition['allocated_size_bytes'] -= lfs['size_bytes']
            else:
                logger_no_user_data.info(f'Keeping LFS "{lfs_uuid}"')
                lfs_uuids_to_keep_in_state.add(lfs_uuid)

        if not dry_run:
            state['large_file_systems'] = {
                key: value for key, value in state['large_file_systems'].items() if key in lfs_uuids_to_keep_in_state
            }

    for path in paths_to_delete:
        logger_no_user_data.info(f'Deleting {path}...')
        if not dry_run:
            subprocess.run(['rm', '-rf', path], check=True)

    logger_no_user_data.info(f'Successfully deleted {len(paths_to_delete)}')



import warnings

from biolib._runtime.runtime import Runtime as _Runtime


def set_main_result_prefix(result_prefix: str) -> None:
    warnings.warn(
        'The "biolib.runtime.set_main_result_prefix" function is deprecated. '
        'It will be removed in future releases from mid 2024. '
        'Please use "from biolib.sdk import Runtime" and then "Runtime.set_main_result_prefix" instead.',
        DeprecationWarning,
        stacklevel=2,
    )
    _Runtime.set_main_result_prefix(result_prefix)



import re
from io import BufferedIOBase, TextIOBase

from biolib.typing_utils import Dict, Iterator, List, Optional, Union


class SeqUtilRecord:
    def __init__(
        self,
        sequence: str,
        sequence_id: str,
        description: Optional['str'] = None,
        properties: Optional[Dict[str, str]] = None,
    ):
        self.sequence = sequence
        self.id = sequence_id  # pylint: disable=invalid-name
        self.description = description

        if properties:
            disallowed_pattern = re.compile(r'[=\[\]\n]')
            for key, value in properties.items():
                assert not bool(disallowed_pattern.search(key)), 'Key cannot contain characters =[] and newline'
                assert not bool(disallowed_pattern.search(value)), 'Value cannot contain characters =[] and newline'
            self.properties = properties
        else:
            self.properties = {}

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} ({self.id})'


class SeqUtil:
    @staticmethod
    def parse_fasta(
        input_file: Union[str, BufferedIOBase, None] = None,
        default_header: Optional[str] = None,
        allow_any_sequence_characters: bool = False,
        use_strict_alphabet: Optional[bool] = False,
        allow_empty_sequence: bool = True,
        file_name: Optional[str] = None,
) -> Iterator[SeqUtilRecord]:
        def process_and_yield_record(header: str, sequence_lines: List[str]):
            sequence = ''.join(sequence_lines)
            sequence_id = header.split()[0]
            if allow_any_sequence_characters and use_strict_alphabet:
                raise Exception(
                    'Error: Please choose either allow_any_sequence_characters or use_strict_alphabet'
                )
            if not allow_any_sequence_characters:
                if use_strict_alphabet:
                    invalid_sequence_characters = SeqUtil._find_invalid_sequence_characters_strict(sequence)
                else:
                    invalid_sequence_characters = SeqUtil._find_invalid_sequence_characters(sequence)
                if invalid_sequence_characters:
                    raise Exception(
                        f'Error: Invalid character ("{invalid_sequence_characters[0]}") found in sequence {sequence_id}'
                    )
            if not allow_empty_sequence and not sequence:
                raise Exception(f'Error: No sequence found for fasta entry {sequence_id}')
            yield SeqUtilRecord(
                sequence=sequence,
                sequence_id=sequence_id,
                description=header[len(sequence_id):].strip()
            )

        def line_generator_from_buffered_io_base(file_handle: BufferedIOBase) -> Iterator[str]:
            for line in file_handle:
                yield line.decode('utf-8')

        def line_generator_from_text_io_base(file_handle: TextIOBase) -> Iterator[str]:
            for line in file_handle:
                yield line

        if input_file is None:
            if file_name:
                input_file = file_name
            else:
                raise ValueError('input_file must be a file name (str) or a BufferedIOBase object')

        file_handle = None
        if isinstance(input_file, str):
            file_handle = open(input_file, "rb")
            line_iterator = line_generator_from_buffered_io_base(file_handle)
        elif isinstance(input_file, BufferedIOBase):
            line_iterator = line_generator_from_buffered_io_base(input_file)
        elif isinstance(input_file, TextIOBase):
            line_iterator = line_generator_from_text_io_base(input_file)
        else:
            raise ValueError('input_file must be a file name (str) or a BufferedIOBase object')

        header = None
        sequence_lines: List[str] = []

        try:
            for line_number, line in enumerate(line_iterator):
                line = line.strip()
                if not line:
                    continue # skip empty lines
                if line.startswith('>'):
                    if header is not None:
                        yield from process_and_yield_record(header, sequence_lines)

                    header = line[1:].strip()
                    sequence_lines = []
                else:
                    if header is None:
                        if default_header:
                            yield from process_and_yield_record(f"{default_header}{line_number}", [line])
                        else:
                            raise Exception(f'No header line found in FASTA file "{file_name}"')
                    else:
                        sequence_lines.append(line)

            if header is not None:
                yield from process_and_yield_record(header, sequence_lines)
        finally:
            if file_handle:
                file_handle.close()

    @staticmethod
    def write_records_to_fasta(file_name: str, records: List[SeqUtilRecord]) -> None:
        with open(file_name, mode='w') as file_handle:
            for record in records:
                optional_description = f' {record.description}' if record.description else ''
                if record.properties:
                    for key, value in record.properties.items():
                        optional_description += f' [{key}={value}]'
                sequence = '\n'.join(record.sequence[i : i + 80] for i in range(0, len(record.sequence), 80))
                file_handle.write(f'>{record.id}{optional_description}\n{sequence}\n')

    @staticmethod
    def _find_invalid_sequence_characters(sequence: str) -> List[str]:
        allowed_sequence_chars = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.')
        invalid_chars = [char for char in sequence if char not in allowed_sequence_chars]
        return invalid_chars

    @staticmethod
    def _find_invalid_sequence_characters_strict(sequence: str) -> List[str]:
        # Equivalent to fair-esm alphabet, compatible with ESM-models
        # Excludes digits, '_' and 'J' (ambiguous letter only used in mass-spec NMR)
        # https://github.com/facebookresearch/esm/blob/2b369911bb5b4b0dda914521b9475cad1656b2ac/esm/constants.py#L8
        allowed_sequence_chars = set('lagvsertidpkqnfymhwcxbuzoLAGVSERTIDPKQNFYMHWCXBUZO-.')
        invalid_chars = [char for char in sequence if char not in allowed_sequence_chars]
        return invalid_chars

    @staticmethod
    def _find_invalid_sequence_id_characters(sequence: str) -> List[str]:
        allowed_chars = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.:*#')
        invalid_chars = [char for char in sequence if char not in allowed_chars]
        return invalid_chars



"""Utilities for displaying file trees and formatting file sizes."""

from typing import Any, Dict, List, Optional, Union

from biolib.biolib_binary_format.utils import LazyLoadedFile


def format_size(size_bytes: int) -> str:
    """Convert bytes to human-readable format (e.g., 2.5KB).

    Args:
        size_bytes: File size in bytes

    Returns:
        Human-readable string representation of the file size
    """
    if size_bytes < 1024:
        return f'{size_bytes}B'
    elif size_bytes < 1024 * 1024:
        return f'{size_bytes/1024:.1f}KB'
    elif size_bytes < 1024 * 1024 * 1024:
        return f'{size_bytes/(1024*1024):.1f}MB'
    else:
        return f'{size_bytes/(1024*1024*1024):.1f}GB'


def build_tree_str(
    data: Dict[str, Any],
    prefix: str = '',
    tree_lines: Optional[List[str]] = None,
    blue: str = '\033[34m',
    white: str = '\033[90m',  # Changed from white (37m) to dark gray (90m) for better visibility on light backgrounds
    reset: str = '\033[0m',
) -> List[str]:
    """Build a string representation of a file tree with color-coded directories and files.

    Args:
        data: Hierarchical tree structure of directories and files
        prefix: Line prefix for indentation and tree structure characters
        tree_lines: List to accumulate tree lines
        blue: ANSI color code for directories
        white: ANSI color code for files (defaults to dark gray for visibility on light backgrounds)
        reset: ANSI color code to reset color

    Returns:
        List of tree lines for display
    """
    if tree_lines is None:
        tree_lines = []

    # Get sorted items, keeping directories first
    items = sorted([(k, v) for k, v in data.items() if k != '__files__'])
    files_list = sorted(data.get('__files__', []), key=lambda f: f['name'])

    # Add directories
    for i, (key, value) in enumerate(items):
        is_last_dir = i == len(items) - 1 and not files_list

        if is_last_dir:
            tree_lines.append(f'{prefix}└── {blue}{key}{reset}')
            build_tree_str(value, prefix + '    ', tree_lines, blue, white, reset)
        else:
            tree_lines.append(f'{prefix}├── {blue}{key}{reset}')
            build_tree_str(value, prefix + '│   ', tree_lines, blue, white, reset)

    # Add files with their sizes
    for i, file in enumerate(files_list):
        is_last = i == len(files_list) - 1
        size_str = format_size(file['size'])
        file_name = file['name']

        if is_last:
            tree_lines.append(f'{prefix}└── {white}{file_name} ({size_str}){reset}')
        else:
            tree_lines.append(f'{prefix}├── {white}{file_name} ({size_str}){reset}')

    return tree_lines


def build_tree_from_files(files: List[LazyLoadedFile]) -> Dict[str, Union[Dict[str, Any], List[Dict[str, Any]]]]:
    """Build a hierarchical tree structure from a list of files.

    Args:
        files: List of files to organize into a tree

    Returns:
        Hierarchical tree structure of directories and files
    """
    tree: Dict[str, Union[Dict[str, Any], List[Dict[str, Any]]]] = {}
    for file in files:
        parts = file.path.lstrip('/').split('/')
        current: Dict[str, Union[Dict[str, Any], List[Dict[str, Any]]]] = tree
        for i, part in enumerate(parts):
            if i == len(parts) - 1:  # This is a file
                if '__files__' not in current:
                    current['__files__'] = []
                files_list = current['__files__']
                assert isinstance(files_list, list)
                files_list.append({'name': part, 'size': file.length})
            else:  # This is a directory
                if part not in current:
                    current[part] = {}
                dir_dict = current[part]
                assert isinstance(dir_dict, dict)
                current = dir_dict
    return tree



from .app import *  # noqa: F403
from .data_record import *  # noqa: F403
from .experiment import *  # noqa: F403
from .resource import *  # noqa: F403
from .resource_permission import *  # noqa: F403
from .resource_version import *  # noqa: F403



import base64
import binascii
import json
import os
from datetime import datetime, timezone
from json.decoder import JSONDecodeError

from biolib._internal.http_client import HttpClient
from biolib._runtime.runtime import Runtime
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.typing_utils import Any, Dict, Optional, TypedDict

from .user_state import UserState


class UserTokens(TypedDict):
    access: str
    refresh: str


class JwtDecodeError(Exception):
    pass


class _ApiClient:
    def __init__(self, base_url: str, access_token: Optional[str] = None):
        self.base_url: str = base_url
        self.access_token: Optional[str] = access_token  # TODO: Deprecate passing access_token in constructor
        self.refresh_token: Optional[str] = None
        self.resource_deploy_key: Optional[str] = None

        self._user_state = UserState()
        self._sign_in_attempted: bool = False

    @property
    def is_signed_in(self) -> bool:
        return bool(self.refresh_token or self.resource_deploy_key)

    def set_user_tokens(self, user_tokens: UserTokens) -> None:
        with self._user_state as user_state:
            user_state['refresh_token'] = user_tokens['refresh']

        self.access_token = user_tokens['access']
        self.refresh_token = user_tokens['refresh']

    def sign_out(self) -> None:
        api_token = os.getenv('BIOLIB_TOKEN', default=None)
        if api_token:
            print('To sign out unset the environment variable "BIOLIB_TOKEN"')

        self.access_token = None
        self.refresh_token = None

        with self._user_state as user_state:
            user_state['refresh_token'] = None

    def refresh_access_token(self) -> None:
        if not self.is_signed_in or self.resource_deploy_key:
            return

        if self.access_token:
            decoded_token = self.decode_jwt_without_checking_signature(self.access_token)
            if datetime.now(tz=timezone.utc).timestamp() < decoded_token['payload']['exp'] - 60:  # 60 second buffer
                # Token has not expired yet
                return

        # TODO: Implement nicer error handling
        try:
            response = HttpClient.request(
                method='POST',
                url=f'{self.base_url}/api/user/token/refresh/',
                data={'refresh': self.refresh_token},
            )
        except Exception as exception:
            logger.error('Sign in with refresh token failed')
            raise exception

        try:
            response_dict = response.json()
        except JSONDecodeError as error:
            logger.error('Could not decode response from server as JSON:')
            raise BioLibError(response.text) from error

        self.access_token = response_dict['access']

    def attempt_sign_in(self) -> None:
        if not self._sign_in_attempted:
            self._attempt_sign_in()
            self._sign_in_attempted = True

    def _attempt_sign_in(self) -> None:
        api_token = os.getenv('BIOLIB_TOKEN', default=None)

        if api_token:
            if api_token.startswith('bld_'):
                self.resource_deploy_key = api_token
            else:
                self.sign_in_with_api_token(api_token)
        else:
            with self._user_state as user_state:
                refresh_token_from_state = user_state['refresh_token']

            # TODO: Handle expired refresh token
            if refresh_token_from_state:
                logger_no_user_data.debug('ApiClient: Signing in with refresh token from user state...')
                self.refresh_token = refresh_token_from_state
                try:
                    self.refresh_access_token()
                except Exception:  # pylint: disable=broad-except
                    self.refresh_token = None
                    with self._user_state as user_state:
                        user_state['refresh_token'] = None

    def sign_in_with_api_token(self, api_token: str) -> None:
        logger_no_user_data.debug('ApiClient: Signing in with BIOLIB_TOKEN...')
        try:
            response = HttpClient.request(
                method='POST',
                url=f'{self.base_url}/api/user/api_tokens/exchange/',
                data={'token': api_token},
            )
        except Exception as exception:
            logger.error('Sign in with API token failed')
            raise exception
        try:
            json_response = response.json()
        except JSONDecodeError as error:
            logger.error('Could not decode response from server as JSON')
            raise BioLibError(response.text) from error

        self.access_token = json_response['access_token']
        self.refresh_token = json_response['refresh_token']

    @staticmethod
    def decode_jwt_without_checking_signature(jwt: str) -> Dict[str, Any]:
        jwt_bytes = jwt.encode('utf-8')

        try:
            signing_input, _ = jwt_bytes.rsplit(b'.', 1)
            header_segment, payload_segment = signing_input.split(b'.', 1)
        except ValueError as error:
            raise JwtDecodeError('Not enough segments') from error

        try:
            header_data = base64.urlsafe_b64decode(header_segment)
        except (TypeError, binascii.Error) as error:
            raise JwtDecodeError('Invalid header padding') from error

        try:
            header = json.loads(header_data)
        except ValueError as error:
            raise JwtDecodeError(f'Invalid header string: {error}') from error

        if not isinstance(header, dict):
            raise JwtDecodeError('Invalid header string: must be a json object')

        try:
            payload_data = base64.urlsafe_b64decode(payload_segment)
        except (TypeError, binascii.Error) as error:
            raise JwtDecodeError('Invalid payload padding') from error

        try:
            payload = json.loads(payload_data)
        except ValueError as error:
            raise JwtDecodeError(f'Invalid payload string: {error}') from error

        if not isinstance(header, dict):
            raise JwtDecodeError('Invalid payload string: must be a json object')

        return dict(header=header, payload=payload)


class BiolibApiClient:
    api_client: Optional[_ApiClient] = None

    @staticmethod
    def initialize(base_url: str, access_token: Optional[str] = None):
        BiolibApiClient.api_client = _ApiClient(base_url=base_url, access_token=access_token)

    @staticmethod
    def get(attempt_sign_in: bool = True) -> _ApiClient:
        api_client = BiolibApiClient.api_client
        if api_client:
            if attempt_sign_in:
                api_client.attempt_sign_in()
            return api_client

        raise BioLibError('Attempted to use uninitialized API client')

    @staticmethod
    def refresh_auth_token():
        api_client = BiolibApiClient.get()
        api_client.refresh_access_token()

    @staticmethod
    def is_reauthentication_needed() -> bool:
        api_client = BiolibApiClient.get()
        if not api_client.is_signed_in and not Runtime.check_is_environment_biolib_app():
            return True
        else:
            return False

    @staticmethod
    def assert_is_signed_in(authenticated_action_description: str) -> None:
        if BiolibApiClient.is_reauthentication_needed():
            raise BioLibError(
                f'You must be signed in to {authenticated_action_description}. '
                f'Please set the environment variable "BIOLIB_TOKEN"'
            )
